/*
 *  Main authors:
 *     Martin Mann <mmann@informatik.uni-freiburg.de>
 *
 *  Contributing authors:
 *     Sebastian Will <will@informatik.uni-freiburg.de>
 *
 *  Copyright:
 *     Martin Mann, 2007
 *
 *  This file is part of the CPSP-tools package:
 *     http://www.bioinf.uni-freiburg.de/sw/cpsp/
 *
 *  See the file "LICENSE" for information on usage and
 *  redistribution of this file, and for a
 *     DISCLAIMER OF ALL WARRANTIES.
 *
 */


#include "HPconvert.hh"

#include <sstream>
#include <fstream>

#include <biu/util/Util_String.h>
#include <cpsp/Exception.hh>

#include <ell/LatticeDescriptorSQR.hh>
#include <ell/LatticeDescriptorCUB.hh>
#include <ell/LatticeDescriptorFCC.hh>


	HPconvert::HPconvert()
	{
	}
	
	HPconvert::~HPconvert()
	{
	}
	
	int
	HPconvert::convert(int argc, char** argv)
	{
		using namespace biu;
		using namespace ell;
		using namespace util;
		
			// the lattice descriptor
		LatticeDescriptor* descr = NULL;
		
		try {
			

			// check user parameters 
			biu::OptionMap allowedArgs;
			std::string infoText;
			initAllowedArguments(allowedArgs,infoText);	// init
			
				// parse programm arguments	
			COptionParser opts = COptionParser(	allowedArgs, argc, 
												(char**)argv, infoText);
				// arguments not parseable or not all mandatory arguments given
			if (!opts.noErrors()) {
				throw cpsp::Exception("",-1);
			}
				// help output
			if (opts.argExist("help")) 
				throw cpsp::Exception("",0);
				// set lattice
			std::string latStr = opts.getStrVal("lat");
			if (latStr.compare("SQR") == 0)
				descr = new LatticeDescriptorSQR();
			else if (latStr.compare("CUB") == 0)
				descr = new LatticeDescriptorCUB();
			else if (latStr.compare("FCC") == 0)
				descr = new LatticeDescriptorFCC();
			else
				throw cpsp::Exception("Unknown lattice type '"+latStr+"'",-2);
			LatticeModel lattice(descr);
				// conversion directions (mode)
			Mode mode = NONE;
			std::string modeStr = opts.getStrVal("m");
			if (modeStr.compare("a2r") == 0) {
				mode = A2R;
			} else if (modeStr.compare("a2p") == 0) {
				mode = A2P;
			} else if (modeStr.compare("r2a") == 0) {
				mode = R2A;
			} else if (modeStr.compare("r2p") == 0) {
				mode = R2P;
			} else if (modeStr.compare("p2a") == 0) {
				mode = P2A;
			} else if (modeStr.compare("p2r") == 0) {
				mode = P2R;
			} else 
				throw cpsp::Exception("Unsupported mode '"+modeStr+"'",-2);
				// check structure input
			std::string structure = opts.getStrVal("s");
			switch (mode) {
				case P2A : // point file 
				case P2R : {
					break;
				} 
				case A2R : // move string
				case A2P :
				case R2A :
				case R2P : {
					if (!lattice.getDescriptor()->getAlphabet()->isAlphabetString(structure))
						throw  cpsp::Exception("Structure '"+structure+"'is no valid move string",-2);
					break;
				} 
				default : assert(false); // should never happen
			}
				// check for verbose output
			bool verboseOut = opts.argExist("v");
			
			if (verboseOut) {
				std::cout	<<"\n==================================="
							<<"\n   HPconvert - CPSP-tool-library"
							<<"\n==================================="
							<<"\n\n"
							<<" conversion of ";
				switch (mode) {
					case A2R : std::cout <<"absolute to relative move string \n\n"; break;
					case A2P : std::cout <<"absolute move string to xyz position file format\n\n"; break;
					case R2A : std::cout <<"relative to absolute move string \n\n"; break;
					case R2P : std::cout <<"relative move string to xyz position file format\n\n"; break;
					case P2A : std::cout <<"xyz positions from file to absolute move string\n\n"; break;
					case P2R : std::cout <<"xyz positions from file to relative move string\n\n"; break;
					default : assert(false); // should never happen
				}
			}
			
				// do conversion
			switch (mode) {
				case A2R : abs2rel(structure, &lattice); break;
				case A2P : abs2xyz(structure, &lattice); break;
				case R2A : rel2abs(structure, &lattice); break;
				case R2P : rel2xyz(structure, &lattice); break;
				case P2A : xyz2abs(structure, &lattice); break;
				case P2R : xyz2rel(structure, &lattice); break;
				default : assert(false); // should never happen
			}
			
			if (verboseOut) {
				std::cout	<<"\n====================================\n"
							<<std::endl;
			}
			
		// catch errors in user parameters etc.
		} catch (cpsp::Exception& ex) {
			std::cerr <<"\n\t"+ex.toString()+"\n\n" <<std::endl;
			return ex.retValue;
		}
		
		if (descr != NULL) delete descr;

		return 0;
	}
	
	void
	HPconvert::abs2rel(	const std::string & absMoves, 
						const ell::LatticeModel * lattice,
						std::ostream & out) const
	{
		out	<<lattice->getString( 
								  lattice->absMovesToRelMoves(
									lattice->parseMoveString(absMoves)
								  )
								) 
			<<std::endl;
	}
		
	void
	HPconvert::rel2abs(	const std::string & relMoves, 
						const ell::LatticeModel * lattice,
						std::ostream & out) const
	{
		out	<<lattice->getString( 
								  lattice->relMovesToAbsMoves(
									lattice->parseMoveString(relMoves)
								  )
								) 
			<<std::endl;
	}
	
	void
	HPconvert::abs2xyz(	const std::string &absMoves,
						const ell::LatticeModel * lattice,
						std::ostream & out) const
	{
			// point data
		ell::IPointVec points = lattice->absMovesToPoints( lattice->parseMoveString(absMoves));
			// write output
		writeXYZ(points, lattice, out);
			// clear data
		points.clear();
	}
	
	void
	HPconvert::rel2xyz(	const std::string &relMoves,
						const ell::LatticeModel * lattice,
						std::ostream & out) const
	{
			// point data
		ell::IPointVec points = lattice->relMovesToPoints( lattice->parseMoveString(relMoves));
			// write output
		writeXYZ(points, lattice, out);
			// clear data
		points.clear();
	}
	
	void
	HPconvert::xyz2abs(	const std::string &xyzFile,
						const ell::LatticeModel * lattice,
						std::ostream & out) const 
	{
			// read point data
		ell::IPointVec points;
		std::ifstream fin(xyzFile.c_str());
		if (fin.good()) {
			readXYZ(fin,points);
		} else {
			std::cout <<"\nERROR: cant read XYZ-file '"+xyzFile+"' !\n" <<std::endl;
			return;
		}
			// convert to absolute move string
		if ( points.size() > 1 ) {
			std::cout	<< lattice->getString( lattice->pointsToAbsMoves(points) )
						<< std::endl;
		} else {
			std::cout <<"\nERROR: only " <<points.size() <<" points read from '"
					<<xyzFile <<"' !\n" <<std::endl;
		}
			// clear data
		fin.close();
		points.clear();
	}
	
	void
	HPconvert::xyz2rel(	const std::string &xyzFile,
						const ell::LatticeModel * lattice,
						std::ostream & out) const 
	{
			// read point data
		ell::IPointVec points;
		std::ifstream fin(xyzFile.c_str());
		if (fin.good()) {
			readXYZ(fin,points);
		} else {
			std::cout <<"\nERROR: cant read XYZ-file '"+xyzFile+"' !\n" <<std::endl;
			return;
		}
			// convert to absolute move string
		if ( points.size() > 1 ) {
			std::cout	<< lattice->getString( lattice->pointsToRelMoves(points) )
						<< std::endl;
		} else {
			std::cout <<"\nERROR: only " <<points.size() <<" points read from '"
					<<xyzFile <<"' !\n" <<std::endl;
		}
			// clear data
		fin.close();
		points.clear();
	}
	
	void 
	HPconvert::writeXYZ(	const ell::IPointVec & points,
							const ell::LatticeModel * lattice,
							std::ostream & out ) const
	{
		out	<<"# lattice protein in absolute positions"
			<<"\n# lattice model  : " <<lattice->getDescriptor()->getName()
			<<"\n# created by HPconvert -- CPSP-tools package"
			<<"\n#"
			<<std::endl;
		for (ell::IPointVec::const_iterator p = points.begin(); p!=points.end(); p++) {
			out <<p->getX() <<" " <<p->getY() <<" " <<p->getZ() <<"\n";
		}
		out	<<"#\n# EOF #" <<std::endl;
		out.flush();
	}
	
	int
	HPconvert::readXYZ(	std::istream & in, ell::IPointVec & points) const
	{
			// clear data
		points.clear();
			// one line of stream and the allowed alphabet for data lines
		std::string line = "", lineAlph = "-0123456789 \t";
			// positions for reading
		int x=0, y=0, z=0;
		while (std::getline(in, line, '\n')) {
				// check if empty line or comment 
			if (line.empty() || (line[0]=='#'))
				continue;
				// check if right alphabet
			else if (line.find_first_not_of(lineAlph,0) == line.npos) {
				std::istringstream iss(line); // feed string stream for conversion
				z = INT_MIN; // init for failure test
				iss >>x >>y >>z; // read absolute lattice position
				if (z != INT_MIN) { // read was successful
					points.push_back(ell::IntPoint(x,y,z)); // write new position
				}
			}
		}
		return 0;
	}

		
	void
	HPconvert::initAllowedArguments(biu::OptionMap & allowedArgs,
											std::string &infoText ) const
	{
		allowedArgs.push_back(biu::COption(	
								"m", false, biu::COption::STRING, 
								"conversion mode (a2r, a2p, r2a, r2p, p2a, p2r) - see information"));
		allowedArgs.push_back(biu::COption(	
								"s", false, biu::COption::STRING, 
								"the structure to convert: either move string or xyz-point file"));
		allowedArgs.push_back(biu::COption(	
								"lat", true, biu::COption::STRING, 
								"lattice (SQR, CUB, FCC)", "CUB"));
		allowedArgs.push_back(biu::COption(	
								"v", true, biu::COption::BOOL, 
								"verbose output"));
		allowedArgs.push_back(biu::COption(	
								"help", true, biu::COption::BOOL, 
								"program parameters and help"));

		infoText = std::string("HPconvert allows the conversion between different")
			+ std::string(" lattice structure representations and prints them")
			+ std::string(" to STDOUT.")
			+ std::string("\nThe supported modes are (m-parameter):\n")
			+ std::string("\n a2r : absolute to relative move string")
			+ std::string("\n a2p : absolute move string to xyz-point file")
			+ std::string("\n r2a : relative to absolute move string")
			+ std::string("\n r2p : relative move string to xyz-point file")
			+ std::string("\n p2a : xyz-point file to absolute move string")
			+ std::string("\n p2r : xyz-point file to relative move string")
			+ std::string("\n\nThe structure is NOT validated")
			+ std::string(" (if connected and selfavoiding). If it is invalid")
			+ std::string(" normal execution cant be guarantied.")
			+ std::string("\n\nThe moves are encoded using:")
			+ std::string("\n F/B : +-x\n L/R : +-y\n U/D : +-z")
			+ std::string("\n\nFor an XYZ-file format example run 'HPconvert")
			+ std::string(" -m=a2p -s=FLUBBURD'. A '#' marks a comment line.");
	}
	



	int main(int argc, char** argv) {
		HPconvert hpc;
			// run conversion
		return hpc.convert(argc, argv);
	}
