/*
 *  Main authors:
 *     Martin Mann <mmann@informatik.uni-freiburg.de>
 *
 *  Contributing authors:
 *     Sebastian Will <will@informatik.uni-freiburg.de>
 *
 *  Copyright:
 *     Martin Mann, 2007
 *
 *  This file is part of the CPSP-tools package:
 *     http://www.bioinf.uni-freiburg.de/sw/cpsp/
 *
 *  See the file "LICENSE" for information on usage and
 *  redistribution of this file, and for a
 *     DISCLAIMER OF ALL WARRANTIES.
 *
 */


#include "HPnnet.hh"
#include <iomanip>
#include <fstream>
#include <algorithm>
#include <queue>

#include <biu/Timer.hh>
#include <biu/util/Util_String.h>

#include <ell/LatticeDescriptorCUB.hh>
#include <ell/LatticeDescriptorFCC.hh>
#include <ell/LatticeModel.hh>
#include <cpsp/Exception.hh>
#include <cpsp/HCoreDatabaseFILE.hh>
#include <cpsp/gecode/GC_HPThreading.hh>

#include <boost/graph/graphviz.hpp>
#include <boost/graph/connected_components.hpp>

///////////////////////////////////////////////////////////////////
// HPconnect CLASS IMPLEMENTATIONS
///////////////////////////////////////////////////////////////////

	HPconnect::HPconnect()
	 : latDescr(NULL), lattice(NULL), coreDB(NULL), minimalOutput(true),
	   alphabetHP("HP",1), stat(NULL)
	{
		clearSequences();
	}
	
	void
	HPconnect::readData(std::istream & input) 
	{
		std::string tmp="", lineBuffer="";
		while (std::getline(input, tmp, '\n')) {
			if (!tmp.empty())
				seqs.push_back(extractSequence(tmp));
		}
	}
	
	std::string
	HPconnect::extractSequence( std::string & str ) const {
		//@TODO extraction of the HP sequence
		return str;
	}
	
	HPconnect::~HPconnect() 
	{
		if (coreDB		!= NULL) { delete coreDB; }
		if (lattice		!= NULL) { delete lattice; }
		if (latDescr	!= NULL) { delete latDescr; }
		if (stat		!= NULL) { delete stat; }
	}
	
	void
	HPconnect::printDot( std::ostream & output ) const {
	    boost::write_graphviz(output, net);
	}	

	void
	HPconnect::printSeqs(std::ostream & output, bool withInd) const {
		for (cpsp::StringVec::size_type s = 0; s < seqs.size(); s++) {
			if (withInd)
				output <<std::setw(8) <<s <<"  ";
			output <<seqs[s] <<std::endl;
		}
	}	    
	
	bool
	HPconnect::areNeighbored( const std::string & s1, const std::string & s2) const {
		std::string a = biu::util::Util_String::invert(s1);
		std::string b = biu::util::Util_String::invert(s2);
		if (a.size() == b.size()) {
			unsigned int differences = 0;
			for (std::string::size_type i = 0; differences < 2 && i<a.size(); i++)
				if (a[i]!=b[i])
					differences++;
			return differences < 2;
		}
		return false;
	}
	
	void
	HPconnect::directConnect( void ) {
		cpsp::StringVec::size_type i,j;
		if (seqs.size() == 0) {
			net = Net();
			return;
		}

//@TODO init nodes of net with all possible indices from seqs
		boost::add_edge(0, seqs.size()-1, net);
		boost::remove_edge(0, seqs.size()-1, net);

		for (i=0; i < seqs.size() ; i++) {
			for (j = i+1; j < seqs.size(); j++) {
				if (areNeighbored(seqs[i], seqs[j])) {
				    boost::add_edge(i, j, net);
				}
			}
		}
	}

	int
	HPconnect::connectedComponents( ComponentLabel & label ) {
		label = ComponentLabel(boost::num_vertices(net),-1);	    
		int labelNumber = boost::connected_components( net, &label[0]);
		return labelNumber;
	}
	
	CSequence
	HPconnect::compressSeq(const std::string& str)
	 const {
	 	using namespace ell;
	 	const Alphabet::Sequence& s = alphabetHP.getSequence(str);
			// base of the number system used for the compression
		size_t base = alphabetHP.getAlphabetSize();
			// the number of positions that can be saved in a byte
		size_t numOfPos = (size_t) (std::log(256.) / std::log((double) base));
			// ceil (s->size() / numOfPos) bytes are necessary to save the
			// sequence
		std::vector<unsigned char> cs(
		 (size_t) std::ceil((double) s.size() / (double) numOfPos), 0);

			// 1 byte = s[0] * base^0 + ... + s[numOfPos] * base^numOfPos
		for (size_t i = 0; i < s.size(); i++) {
			cs[i/numOfPos] +=
			 alphabetHP.getIndex(s.at(i)) * (int) pow(base, i%numOfPos);
		}
		return cs;
	}

	std::string
	HPconnect::uncompressSeq(const std::vector<unsigned char>& cs,
		size_t seqLength) const 
	{
			// base of the number system used for the compression
		size_t base = alphabetHP.getAlphabetSize();
			// the number of positions that can be saved in a byte
		size_t numOfPos = (size_t) (std::log(256.) / std::log((double) base));
		std::string uncompressedSeq = "";
			// element of cs which is currently processed
		unsigned char c = cs[0];

		assertell(seqLength <= cs.size()*numOfPos,
		 "The compressed sequence can't hold seqLength many elements.");

			// 1 byte = s[0] * base^0 + ... + s[numOfPos-1] * base^(numOfPos-1)
		for (size_t i = 0; i < seqLength; i++) {
				// update c if new element of cs is processed
			if (i%numOfPos == 0) {
				c = cs[i/numOfPos];
			}
			uncompressedSeq += alphabetHP.getString(alphabetHP.getElement(
								c % base));
			c /= base;
		}

		return uncompressedSeq;
	}
	
	
	bool
	HPconnect::initThreading(int argc, char** argv, cpsp::HPThreadingHandler & t) {
		
		cpsp::gecode::GC_HPThreading & threading = (cpsp::gecode::GC_HPThreading &)t;
		
		biu::OptionMap allowedArgs;
		std::string infoText;
		initAllowedArguments(allowedArgs,infoText);	// init

			// parse programm arguments
		biu::COptionParser opts = biu::COptionParser(	allowedArgs, argc,
														(char**)argv, infoText);
			// arguments parseable and all mandatory arguments given
		if (opts.noErrors()) {
				// help output
			if (opts.getBoolVal("help")) {
				opts.coutUsage();
				return false;
			}
				// set run mode
			threading.setRunMode(cpsp::gecode::GC_HPThreading::ENUMERATE);

				// verbose output
			threading.setStatOutput(false);
			minimalOutput = opts.getBoolVal("s");
			threading.setVerboseOutput(false);
			threading.setNoOutput(minimalOutput);
				// using symmetryBreaking
			threading.setSymmetryBreaking(! opts.getBoolVal("noSymmBreak"));
				// using DDS for counting only
			threading.setUsingDDS(false); // no DDS so far for enumeration

			threading.setNoRecomputation(opts.getBoolVal("noRecomp"));
				// core selection mode
			threading.setCoreSelection(cpsp::gecode::GC_HPThreading::ALL_BEST);
			// maximal structure number to take into find
			int tmp = opts.getIntVal("maxDeg");
			if (tmp < 1) {
				std::cerr <<"Error in arguments : maxSol <= 0\n";
				return false;
			}
			threading.setMaxStructures((unsigned int) tmp+1);
			maxDeg = (unsigned int)tmp;

				// sequence
			std::string s = opts.getStrVal("seq");
			for (std::string::iterator si = s.begin(); si!=s.end(); si++) {
				if (*si == 'h') 		{ *si = 'H';
				} else if (*si == 'p')	{ *si = 'P';
				} else if (*si != 'H' && *si != 'P') {
					std::cerr <<"Error in arguments: sequence is no HP-sequence : " 
								<<s <<"\n";
					return false;
				}
			}
			threading.setSequence(s);

				// verbose output if not silent mode choosen
			verboseOut = !opts.argExist("s");

				// lattice modell
			if (lattice != NULL) {
				delete lattice;
				lattice = NULL;
			}
			if (latDescr != NULL) { 
				delete latDescr; 
				latDescr = NULL;
			}
				
			if (opts.getStrVal("lat") == "CUB")
				latDescr = new ell::LatticeDescriptorCUB();
			else if (opts.getStrVal("lat") == "FCC")
				latDescr = new ell::LatticeDescriptorFCC();
			else {
				std::cerr <<"Error in arguments: lattice is not supported\n";
				return false;
			}
			threading.setLatticeDescriptor(latDescr);
			lattice = new ell::LatticeModel(latDescr);
			
				// structure
			structGiven = opts.argExist("struct");
			structsRel.clear();
			if (structGiven) {
				std::string str = opts.getStrVal("struct");
				if (str.size() != (s.size()-1)) {
					std::cerr <<"Error in arguments: structure and sequence are of"
							<<" different sizes\n";
					return false;
				}
				if (!latDescr->getAlphabet()->isAlphabetString(str)) {
					std::cerr <<"Error in arguments: structure is no allowed"
							<<" Absative move string in this lattice\n";
					return false;
				}
				
				ell::IPointVec strPoints(lattice->absMovesToPoints(
										lattice->parseMoveString(
											str )));

				const ell::AutomorphismVec autM = lattice->getDescriptor()->getAutomorphisms();
				ell::AutomorphismVec::const_iterator a;
				ell::IPointVec act = strPoints;
				for (a = autM.begin(); a!= autM.end(); a++) {
					for (ell::IPointVec::size_type pos=0; pos < strPoints.size(); pos++) {
						act[pos] = (*a) * strPoints[pos];
					}
					structsRel.insert(lattice->getString(lattice->pointsToRelMoves(act)));
				}
				
				if(verboseOut) {
					std::cout <<"\n symmetric structures\n";
					for (StringSet::const_iterator it = structsRel.begin(); it != structsRel.end(); it++)
						std::cout <<" rel = " <<(*it) <<"\n";
				}
			}

				// energie limits
			int minE = INT_MIN/2+2;
			int maxE = 0;
// @TODO option wieder reinnehmen
//			minE = opts.getIntVal("minE");
//			maxE = opts.getIntVal("maxE");
				// semantische fehler pruefen
			if (minE > maxE)
				throw cpsp::Exception("Error in arguments : minE > maxE", -2);

				// contact limits for database init
			unsigned int minContacts = -maxE + threading.getContactsInSeq();
			unsigned int maxContacts = -minE + threading.getContactsInSeq();

				// H-core database
			char * tmpStr = NULL;
			if (coreDB != NULL) { delete coreDB; coreDB = NULL;}
			if (opts.argExist("dbPath")) {
				coreDB = new cpsp::HCoreDatabaseFILE(opts.getStrVal("dbPath"));
			} else if ((tmpStr = getenv("CPSP_COREDB")) != NULL) {
				coreDB = new cpsp::HCoreDatabaseFILE(std::string(tmpStr));
			} else {
				std::cerr << "H-core database Error: no database available (FILE)\n";
			}
				// testen
			if (!coreDB->isConnected()) {
				std::cerr << "H-core database Error: can't open database\n";
				return false;
			}
			if (!coreDB->initCoreAccess(	*latDescr,
											threading.getCoreSize(),
											minContacts,
											maxContacts)) {
				std::cerr <<"H-core database Error: access init failed\n";
				return false;
			}
			threading.setCoreDB(coreDB);
			
			doLexOrdering = opts.argExist("lexOrd");
			
			if (opts.argExist("stat")) {
				if (stat != NULL) delete stat;
				stat = new Statistic();
			}
			
			return true;
		} 
		std::cerr << "Parameter Error: not parseable\n";
		return false;
	}
	
	bool
	HPconnect::initNextThreading(const std::string & seq,
		cpsp::HPThreadingHandler & t) 
	{
		assert(coreDB != NULL);
		assert(latDescr != NULL);
		
		cpsp::gecode::GC_HPThreading & threading = (cpsp::gecode::GC_HPThreading &)t;
		
			// next sequence
		threading.setSequence(seq);
		
			// reinit core db 
		if (!coreDB->initCoreAccess(	*latDescr,
										threading.getCoreSize(),
										threading.getContactsInSeq(),
										UINT_MAX/2)) {
			std::cerr <<"H-core database Error: access init failed\n";
			return false;
		}
		threading.setCoreDB(coreDB);
		
		return  true;
	}


	void
	HPconnect::initAllowedArguments(biu::OptionMap & allowedArgs,
											std::string &infoText ) const {
		allowedArgs.push_back(biu::COption(
								"seq", false, biu::COption::STRING,
								"HP-sequence to expand"));
		allowedArgs.push_back(biu::COption(
								"struct", true, biu::COption::STRING,
								"structure all sequences have to fold in (abs. Moves)"));
		allowedArgs.push_back(biu::COption(
								"dbPath", true, biu::COption::STRING,
								"file based database root path (or use $CPSP_COREDB)"));
		allowedArgs.push_back(biu::COption(
								"lat", true, biu::COption::STRING,
								"lattice (CUB, FCC)", "CUB"));
		allowedArgs.push_back(biu::COption(
								"noSymmBreak", true, biu::COption::BOOL,
								"do not break symmetries"));
		allowedArgs.push_back(biu::COption(
								"maxDeg", true, biu::COption::INT,
							"maximal degeneration of the sequences (value > 0)",
								biu::util::Util_String::int2str(1)));
		allowedArgs.push_back(biu::COption(
								"noRecomp", true, biu::COption::BOOL,
								"no space recomputation"));
		allowedArgs.push_back(biu::COption(
								"s", true, biu::COption::BOOL,
								"silent - minimal output"));
		allowedArgs.push_back(biu::COption(
								"stat", true, biu::COption::BOOL,
								"print run statistics"));
		allowedArgs.push_back(biu::COption(
								"lexOrd", true, biu::COption::BOOL,
								"use always lexicographically smaller of sequence and reversion"));
		allowedArgs.push_back(biu::COption(
								"help", true, biu::COption::BOOL,
								"program parameters and help"));

		infoText = std::string("HPnnet calculates the neutral net for a given ")
				+ std::string("sequence and structure. All sequences have to have ")
				+ std::string("a degeneracy below a given level, have to form ")
				+ std::string("the given structure as an optimal, and to be a ")
				+ std::string("direct or indirect neighbor of the initial sequence ")
				+ std::string("by mutating sequence positions.");

	} // initArguments
	
	
	bool
	HPconnect::structFound(const cpsp::StringVec & sols) const {
		assert(lattice != NULL);
		if (!structGiven)
			return true;
		bool found = false;
		for (cpsp::StringVec::const_iterator i = sols.begin(); !found && i!=sols.end();i++) {
			if (i->compare("") == 0) continue;			
			 // get relative move string
			std::string str = lattice->getString(
								lattice->absMovesToRelMoves(
									lattice->getDescriptor()->getSequence(*i)));
			 // search if one of the symmetrics to the initial one
			found = std::find(structsRel.begin(),structsRel.end(),str) != structsRel.end();
		}
		return found;
	}
	
	cpsp::StringVec
	HPconnect::processQueue(cpsp::HPThreadingHandler & t, SeqQueue & q, unsigned int seqSize) {

		cpsp::gecode::GC_HPThreading & threading = (cpsp::gecode::GC_HPThreading &)t;

		std::string actSeq = "", tmp="";
		std::string * next = &actSeq;
		std::string::size_type i = 0;
		CSequence cseq;
		unsigned int deg = 0;
		cpsp::StringVec sols, otherNets;
		while(!q.empty()) {
			actSeq = uncompressSeq(q.front(),seqSize);
			std::cout <<" "<<std::setw(4) <<seqs.size() <<"  "<<actSeq <<" " <<seq2deg[q.front()] <<std::endl;
			seqs.push_back(actSeq);
			q.pop();
			for (i = 0; i< actSeq.size(); i++) {
					// flip position i
				if (actSeq[i] == 'H') actSeq[i] = 'P';
				else actSeq[i] = 'H';
				
				if (doLexOrdering) {
					tmp = actSeq;
						// tmp == reverted sequence of actSeq
					std::reverse_copy(actSeq.begin(),actSeq.end(),tmp.begin());
	
						// reference to lexicographically smaller sequence				
					next = (tmp < actSeq) ? &tmp : &actSeq;
				} else {
					next = &actSeq;
				}

				cseq = compressSeq(*next);
					// if not already processed
				if (seq2deg.find(cseq)==seq2deg.end()) {
						// init threading
					initNextThreading(*next, threading);
					sols.clear();	
						// enumerate optimal structures by threading
					deg = threading.enumerateStructures(sols);
						// mark as seen
					seq2deg[cseq] = deg;
	
//					std::cout <<"\n new " <<(*next) <<" with deg = " <<deg <<"";
					 // if degeneracy low enough and able to build searched structure
					if (deg > 0 && deg <= maxDeg) {
						if (verboseOut) {
							std::cout <<std::setw(5) <<" " <<"  --> " <<(*next) <<" added"<<std::endl;
						}
						 // add to queue
						if( structFound(sols)) {
							q.push(cseq);
						} else {
							otherNets.push_back(*next);
						}
					}
				} else {
//					std::cout <<"\n    " <<(*next) <<" already seen";
				}
				
								
					// flip position i back
				if (actSeq[i] == 'H') actSeq[i] = 'P';
				else actSeq[i] = 'H';
			}
			
		}
		return otherNets;
	}

	int
	HPconnect::run(int argc, char** argv) {
		seq2deg.clear();		// clear hash of seen sequences
		cpsp::StringVec sols;	// holds generated structures

		cpsp::gecode::GC_HPThreading threading;	// object to determine degeneracy

			// timer for statistics
		biu::Timer timer; timer.start();
		
		if (initThreading(argc,argv,threading)) {	// init was ok
			// initial test
			std::string seq = threading.getSequence();
			if (doLexOrdering) {
				std::string tmp = seq;
				std::reverse(tmp.begin(),tmp.end());
				if ( tmp < seq ) {
					std::cerr <<"\n\tError: " 
							<<"Please give reverted sequence and corresponding structure.\n\n";
					return -1;
				}
			}
			unsigned int deg = threading.enumerateStructures(sols);
			if (deg > maxDeg) {
				std::cerr <<"\n\tError: " 
						<<"Degeneracy of initial sequence higher than allowed (deg. > "
						<< maxDeg <<" ).\n\n";
				return -1;
			}
			if (!structFound(sols)) {
				std::cerr <<"\n\tError: " 
						<<"Given structure is not optimal for initial sequence.\n\n";
				for (size_t x = 0; x<sols.size(); x++)  {
					std::cout <<" - "<<sols[x] <<"\n";
				}
				return -1;
			}
			
				// init queue of compressed HP-sequences to check
			CSequence cseq = compressSeq(seq);
			SeqQueue toCheck;
			toCheck.push(cseq);
			seq2deg[cseq] = deg;
				// sequences of this neutral net
			seqs.clear();
			
				// process queue until empty
			std::cout	<<"\n###################################################"
						<<"\n##  the neutral net (ID,SEQUENCE,DEGENERACY) :  ###"
						<<"\n###################################################"
						<<"\n" <<std::endl;
			cpsp::StringVec otherNets = processQueue(threading, toCheck, seq.size());
			
			if (stat != NULL) {
				stat->hashEntries = seq2deg.size();
			}
			
				// print network
			std::cout	<<"\n###################################################"
						<<"\n##  the neutral net in DOT format :  ##############"
						<<"\n###################################################"
						<<"\n" <<std::endl;
			directConnect();
			printDot(std::cout);
			
				// give remaining nets with different structure
			if (otherNets.size() > 0) {
				std::cout	<<"\n###################################################"
							<<"\n##  sequences of other neutral nets :  ############"
							<<"\n###################################################"
							<<"\n" <<std::endl;
				for (cpsp::StringVec::const_iterator it = otherNets.begin(); it != otherNets.end(); it++) {
					std::cout <<std::setw(7) <<" "<<(*it)<<" " <<seq2deg[compressSeq(*it)]<<"\n";
				}
				std::cout.flush();
			}
			
		} else {
			return -1;
		}
		
		if (stat != NULL) {
			stat->time = timer.stop();
			std::cout	<<"\n###################################################"
						<<"\n##  statistics :  #################################"
						<<"\n###################################################"
						<<"\n"
						<<"\n    runtime      : " <<stat->time <<" ms"
						<<"\n    hashed Seqs. : " <<stat->hashEntries
						<<"\n" <<std::endl; 
		}
		
		std::cout	<<"\n###################################################"
					<<"\n" <<std::endl; 
		return 0;
	}


///////////////////////////////////////////////////////////////////
// THE MAIN FUNCTION
///////////////////////////////////////////////////////////////////

	int main(int argc, char** argv) {
		HPconnect hpConnect;
		
/*		std::ifstream input;
		input.open("test.seq");
		if (input.good()) {
			hpConnect.readData(input);
			input.close();
		} else {
			std::cerr <<"could not open file 'test.seq'\n";
		}
		
		hpConnect.directConnect();
			std::cout <<std::endl;
		hpConnect.printSeqs(std::cout);
			std::cout <<std::endl;
		hpConnect.printDot(std::cout);
			std::cout <<std::endl;

		HPconnect::ComponentLabel label;
		std::cout	<<"connected components = " 
					<<hpConnect.connectedComponents(label)
					<<std::endl;
		for (HPconnect::ComponentLabel::size_type i = 0; i< label.size(); i++) {
			std::cout <<std::setw(8) <<i <<" : " <<std::setw(4) <<label[i] <<std::endl;
		}
		std::cout <<std::endl;
*/


		return hpConnect.run(argc,argv);
	}
